#define F_CPU 16000000

#include <avr/io.h>
#include <util/delay.h>
#include "usart.h"
#include <avr/interrupt.h>
#include "string.h"

/*
 Driver fata: PD2-PD5, enable PD6-PD7 
 Driver spate: PB0-PB2 si PB5, enable PB3-PB4
 Infrarosu: Fata stanga PA0, fata dreapta PA1, spate-stanga PA2, spate-dreapta PA3
 Ultrasunete: fata PC0-1, fata-stanga PC2-3, fata-dreapta PC4-5, stanga PC6-PC7, dreapta PB6-7
*/

char buffer[1024];
int volatile v[4];
int volatile lastWheel[4];
int volatile measure_ctr;
int volatile is_measuring[5], ultrasound_start[5], distance[5], has_measured[5];

void measure_distances() {
	PORTC = 0;
	_delay_us(2);
	DDRC = 1;
	PORTC = (1 << PC0); // | (1 << PC2) | (1 << PC4) | (1 << PC6);
	_delay_us(10);
	has_measured[0] = has_measured[1] = has_measured[2] = has_measured[3] = 0;
	PORTC = 0;
	DDRC = 0;
	PORTC = (1 << PC1); // | (1 << PC3) | (1 << PC5) | (1 << PC7);
}

ISR(TIMER1_COMPA_vect)
{
	sprintf(buffer, "U0=%d,I0=%d,I1=%d,I2=%d,I3=%d|",
		distance[0],
		v[0], v[1], v[2], v[3]);
	USART0_print(buffer);
	measure_ctr = (measure_ctr + 1) % 3;
	if (measure_ctr == 0)
		measure_distances();
}

ISR(PCINT0_vect)
{
	if ((PINA & (1 << PA0)) >> PA0 != lastWheel[0]) {
		++v[0];
		lastWheel[0] = 1 - lastWheel[0];
	}
	if ((PINA & (1 << PA1)) >> PA1 != lastWheel[1]) {
		++v[1];
		lastWheel[1] = 1 - lastWheel[1];
	}
	if ((PINA & (1 << PA2)) >> PA2 != lastWheel[2]) {
		++v[2];
		lastWheel[2] = 1 - lastWheel[2];
	}
	if ((PINA & (1 << PA3)) >> PA3 != lastWheel[3]) {
		++v[3];
		lastWheel[3] = 1 - lastWheel[3];
	}
}

ISR(PCINT2_vect)
{
	// ONE TIMER TICK = 16 us (equivalent to 5.4 mm) =>
	if (!has_measured[0]) {
		if ((PINC & (1 << PC1)) != 0 && !is_measuring[0]) {
			ultrasound_start[0] = TCNT1;
			is_measuring[0] = 1;
		} else if ((PINC & (1 << PC1)) == 0 && is_measuring[0]) {
			is_measuring[0] = 0;
			distance[0] = TCNT1 - ultrasound_start[0];
			if (distance[0] < 0) distance[0] += 6250;
			distance[0] = distance[0] / 4;
			has_measured[0] = 1;
		}
	}

/*	if (!has_measured[1]) {
		if ((PINC & (1 << PC3)) != 0 && !is_measuring[1]) {
			ultrasound_start[1] = TCNT1;
			is_measuring[1] = 1;
		} else if ((PINC & (1 << PC3)) == 0 && is_measuring[1]) {
			is_measuring[1] = 0;
			distance[1] = TCNT1 - ultrasound_start[1];
			if (distance[1] < 0) distance[1] += OCR1A;
			distance[1] = distance[1] * 528 / 1000;
			has_measured[1] = 1;
		}
	}

	if (!has_measured[2]) {
		if ((PINC & (1 << PC5)) != 0 && !is_measuring[2]) {
			ultrasound_start[2] = TCNT1;
			is_measuring[2] = 1;
		} else if ((PINC & (1 << PC5)) == 0 && is_measuring[2]) {
			is_measuring[2] = 0;
			distance[2] = TCNT1 - ultrasound_start[2];
			if (distance[2] < 0) distance[2] += OCR1A;
			distance[2] = distance[2] * 528 / 1000;
			has_measured[2] = 1;
		}
	}

	if (!has_measured[3]) {
		if ((PINC & (1 << PC7)) != 0 && !is_measuring[3]) {
			ultrasound_start[3] = TCNT1;
			is_measuring[3] = 1;
		} else if ((PINC & (1 << PC7)) == 0 && is_measuring[3]) {
			is_measuring[3] = 0;
			distance[3] = TCNT1 - ultrasound_start[3];
			if (distance[3] < 0) distance[3] += OCR1A;
			distance[3] = distance[3] * 528 / 1000;
			has_measured[3] = 1;
		}
	} */
}

void init_engines()
{
	// Front driver
	DDRD = 0;
	PORTD = 0;
	DDRD |= (1 << PD2) | (1 << PD3) | (1 << PD4) | (1 << PD5);
	DDRD |= (1 << PD6) | (1 << PD7); // OC2
	
	// Back driver
	DDRB = 0;
	PORTB = 0;
	DDRB |= (1 << PB0) | (1 << PB1) | (1 << PB2) | (1 << PB5);
	DDRB |= (1 << PB3) | (1 << PB4); // OC0


	// Set PWM for front (values 150-255)
	TCCR2A = 0;
	TCCR2B = 0;

	TCCR2A |= (3 << WGM20);
	OCR2A = OCR2B = 190;

	TCCR2A |= (2 << COM2A0);
	TCCR2A |= (2 << COM2B0);

	TCCR2B |= (3 << CS20);

	// Set PWM for back (values 150-255)
	TCCR0A = 0;
	TCCR0B = 0;

	TCCR0A |= (3 << WGM00);
	OCR0A = OCR0B = 190;

	TCCR0A |= (2 << COM0A0);
	TCCR0A |= (2 << COM0B0);

	TCCR0B |= (3 << CS00);
}

void init_sensors()
{
	// Ports for wheel spin sensors
	DDRA = 0;
	PORTA = (1 << PA0) | (1 << PA1) | (1 << PA2) | (1 << PA3);
	PCICR = (1 << PCIE0);
	PCMSK0 = 15;

	// Initialize wheel spin variables
	v[0] = v[1] = v[2] = v[3] = 0;
	lastWheel[0] = lastWheel[1] = lastWheel[2] = lastWheel[3] = 0;
	if ((PINA & (1 << PA0)) == 0) lastWheel[0] = 1;
	if ((PINA & (1 << PA1)) == 0) lastWheel[1] = 1;
	if ((PINA & (1 << PA2)) == 0) lastWheel[2] = 1;
	if ((PINA & (1 << PA3)) == 0) lastWheel[3] = 1;
	

	// Initialize for ultrasound distance sensors
	measure_ctr = 0;
	is_measuring[0] = is_measuring[1] = is_measuring[2] = is_measuring[3] = 0;
	has_measured[0] = has_measured[1] = has_measured[2] = has_measured[3] = 1;
	PCICR |= (1 << PCIE2);
	PCMSK2 = (1 << PCINT17) | (1 << PCINT19) | (1 << PCINT21) | (1 << PCINT23);

	// Initialize timer 1 (bluetooth send interrupts) - 100ms
	// CARE IF THIS IS CHANGED (USED FOR ULTRASOUND TOO)
	//TCCR1B = (1 << WGM12) | (1 << CS11) | (1 << CS10);
	//OCR1A = 0x61A8;
	TCCR1A = 0;
	TCCR1B = (1 << WGM12) | (1 << CS12);
	OCR1A = 6250;
	TIMSK1 |= (1 << OCIE1A);

	sei();
}


void move_forward()
{
	// Front
	PORTD |= (1 << PD3);
	PORTD &= ~(1 << PD2);
	PORTD &= ~(1 << PD5);
	PORTD |= (1 << PD4);

	// Back
	PORTB |= (1 << PB0);
	PORTB &= ~(1 << PB1);
	PORTB &= ~(1 << PB2);
	PORTB |= (1 << PB5);
}

void move_backwards()
{
	// Front
	PORTD |= (1 << PD2);
	PORTD &= ~(1 << PD3);
	PORTD &= ~(1 << PD4);
	PORTD |= (1 << PD5);

	// Back
	PORTB |= (1 << PB1);
	PORTB &= ~(1 << PB0);
	PORTB &= ~(1 << PB5);
	PORTB |= (1 << PB2);
}

void rotate_in_place1() {
	// Front
	PORTD |= (1 << PD2);
	PORTD &= ~(1 << PD3);
	PORTD &= ~(1 << PD5);
	PORTD |= (1 << PD4);

	// Back
	PORTB |= (1 << PB0);
	PORTB &= ~(1 << PB1);
	PORTB &= ~(1 << PB5);
	PORTB |= (1 << PB2);
}

void rotate_in_place2() {
	// Front
	PORTD |= (1 << PD3);
	PORTD &= ~(1 << PD2);
	PORTD &= ~(1 << PD4);
	PORTD |= (1 << PD5);

	// Back
	PORTB |= (1 << PB1);
	PORTB &= ~(1 << PB0);
	PORTB &= ~(1 << PB2);
	PORTB |= (1 << PB5);
}

void stop() {
	PORTD &= ~((1 << PD2) | (1 << PD3) | (1 << PD4) | (1 << PD5));
	PORTB &= ~((1 << PB0) | (1 << PB1) | (1 << PB2) | (1 << PB5));
}

int main() {
	init_engines();
	USART0_init();
	init_sensors();

	char c;
	while (1) {
		
		c = USART0_receive();
		if (c == 'W') {
			move_forward();
    		} else if (c == 'S') {
			move_backwards();
		} else if (c == 'L') {
			rotate_in_place1();
		} else if (c == 'R') {
			rotate_in_place2();
		} else if (c == 'Q') {
			stop();
		}
		else if (c == 'z') {
			// Received new speed
			int nr = 0;
			c = USART0_receive();
			nr = nr * 10 + (c - '0');
			c = USART0_receive();
			nr = nr * 10 + (c - '0');
			c = USART0_receive();
			nr = nr * 10 + (c - '0');
			OCR0A = OCR0B = OCR2A = OCR2B = nr;
		}
	}

	return 0;
}
